"""
    Exception & Error Messages
"""


class MissingValue(Exception):
    """Base class for other exceptions"""


def error_message(message):
    """Error Message"""

    outlined = f"{ '~' * (len(message) + 4)}"
    return f"\n{outlined}\n~ {message} ~\n{outlined}"



"""
    Tool for TOML (before tomlib gets implemented)
"""

import toml

TOML_SETUP = """
[spoc]
name = "my-project"
version = "0.1.0"

[spoc.config]
mode = "development" # development, production, staging, custom

[spoc.apps]
production = []
development = []
staging = []
""".strip()


class TOML:
    """TOML Wrapper"""

    file = "spoc.toml"
    setup = TOML_SETUP

    @classmethod
    def read(cls, file_to_read = None):
        """Read"""
        file = file_to_read or cls.file
        with open(file, "r", encoding="utf-8") as found_file:
            toml_string = found_file.read()
            parsed_toml = toml.loads(toml_string)
        return parsed_toml

    @classmethod
    def init(cls):
        """Initialize"""
        with open(cls.file, "w", encoding="utf-8") as file:
            file.write(cls.setup)



"""
    Module(INIT)
"""

import pathlib

from .framework import framework
from .frozendict import FrozenDict as frozendict
from .components import component
from .project import project
from .tools import get_attr, get_fields
from .imports import search_method

def root(path):
    """Easy Level-Up Folder(s)."""

    return pathlib.Path(path).parents



"""
    DataClasses
"""

import dataclasses as dc
import typing


@dc.dataclass(frozen=True)
class Class:
    """Project Classes"""

    name: str = None
    app: str = None
    module: str = None
    key: str = None
    uri: str = None
    cls: typing.Any = None


@dc.dataclass(frozen=True)
class Definition:
    """Project Definitions"""

    path: str
    app: str
    module: str
    fields: dict[str, typing.Any]


@dc.dataclass(frozen=True)
class API:
    """Project API"""

    modules: dict
    plugins: dict


@dc.dataclass(frozen=True)
class Component:
    """Project Plugin"""

    config: typing.Any = None
    metadata: typing.Any = None
    is_spoc_plugin: bool = True


@dc.dataclass(frozen=True)
class Global:
    """Project Globals"""

    schema: typing.Any = None
    modules: typing.Any = None


@dc.dataclass(frozen=True)
class Project:
    """Project Core"""

    keys: typing.Any
    modules: typing.Any
    schema: typing.Any
    components: typing.Any
    settings: typing.Any
    installed_apps: list[str]
    toml: typing.Any
    pyproject: typing.Any



"""

___________                            __      __             __    
\_   _____/___________    _____   ____/  \    /  \___________|  | __
 |    __) \_  __ \__  \  /     \_/ __ \   \/\/   /  _ \_  __ \  |/ /
 |     \   |  | \// __ \|  Y Y  \  ___/\        (  <_> )  | \/    < 
 \___  /   |__|  (____  /__|_|  /\___  >\__/\  / \____/|__|  |__|_ \
     \/               \/      \/     \/      \/                   \/

"""

import typing

import spoc


@spoc.framework
class Framework:
    """Framework Builder"""

    plugins = ["types", "graphql", "router", "commands"]


@spoc.project
class Project:
    """Framework"""

    def init(
        self,
        base_dir: typing.Any = None,
        mode: str = "cli",
    ):
        """Class __init__ Replacement"""

        # Step[1]: INIT { Admin }
        Framework(base_dir=base_dir, mode=mode, app=self)

        # Finally: Collect { Keys }
        self.keys = [x for x in spoc.get_fields(self) if x not in ["init"]]



"""
    Framework
"""

import functools
import os
import pathlib
import sys

from .errors import MissingValue, error_message
from .frozendict import FrozenDict
from .get_spoc import get_spoc_plugins
from .imports import get_plugins
from .toml_app import TOML
from .tools import get_fields
from .types import Project

CORE_KEYS = [
    "app_mode",
    "apps",
    "base_dir",
    "installed_apps",
    "mode",
    "modules",
    "plugins",
    "project",
    "schema",
    "settings",
    "toml",
]
PROJECT_KEYS = [
    "modules",
    "schema",
    "components",
    "settings",
    "installed_apps",
    "toml",
    "pyproject",
    "keys",
]


class FrameworkSingleton:
    """Class: Create a Singleton"""

    def __new__(cls, *args, **kwargs):
        it_id = "__it__"
        it = cls.__dict__.get(it_id, None)
        if it is not None:
            return it
        it = object.__new__(cls)
        setattr(cls, it_id, it)
        it.init(cls, *args, **kwargs)
        return it

    def init(self, *args, **kwargs):
        """Class __init__ Replacement"""


def create_project(self, cls, base_dir):
    """Wraps The Project Configurations"""
    # Step[1]: INIT { Admin }
    admin = self

    # Step[2]: Self { Definitions }
    cls.admin = admin
    cls.base_dir = base_dir
    cls.mode = admin.app_mode
    cls.project = admin.project
    core_plugins = admin.plugins

    # Cleaned Modules (IF: Founded)
    core_modules = None
    if core_plugins:
        core_modules = core_plugins.modules

    cls.core = Project(
        toml=self.toml,
        installed_apps=admin.installed_apps,
        schema=admin.schema,
        modules=core_modules,
        components=admin.modules,
        settings=admin.settings,
        keys=[x for x in PROJECT_KEYS if x not in ["keys"]],
        pyproject=admin.pyproject,
    )
    return cls


def collect_installed_apps(self):
    """Collect All Apps"""
    # Step[1]: INIT { Values }
    admin = self
    app_mode = admin.app_mode
    settings = admin.settings
    apps = admin.apps
    installed_apps = []

    # Step[2]: Collect Apps
    match app_mode:
        case "custom":
            if hasattr(settings, "INSTALLED_APPS"):
                installed_apps.extend(settings.INSTALLED_APPS)
        case "production":
            installed_apps.extend(apps.get("production", []))
        case "staging":  # production + staging
            installed_apps.extend(apps.get("production", []))
            installed_apps.extend(apps.get("staging", []))
        case "development":  # production + development
            installed_apps.extend(apps.get("production", []))
            installed_apps.extend(apps.get("development", []))

    self.load_apps(self, installed_apps=installed_apps)
    return installed_apps


def init(
    self,
    base_dir: pathlib.Path = None,
    mode: str = None,
    plugins: list[str] = None,
    app: object = None,
):
    """__init__ Only Runs Once Per Settings Project."""

    if not base_dir:
        raise MissingValue(error_message("Project(base_dir = pathlib.Path)"))

    # Step[1]: INIT { Definitions }
    disabled_fields = ["mode", "base_dir"]
    for field in CORE_KEYS:
        if field not in disabled_fields:
            setattr(self, field, None)

    # Step[2]: Self { Definitions }
    self.mode = mode
    self.base_dir = base_dir
    self.__plugins__ = plugins
    self.__theclass__ = app

    # Step[3]: Inject { Apps }
    base_apps = pathlib.Path(base_dir / "apps")
    base_apps.mkdir(parents=True, exist_ok=True)
    if base_apps.exists():
        sys.path.insert(0, os.path.join(base_dir, "apps"))

    # Step[4.1]: TOML { Config }
    toml_file = pathlib.Path(base_dir / "spoc.toml")
    pytoml_file = pathlib.Path(base_dir / "pyproject.toml")

    # Step[4.2]: Load { TOML }
    TOML.file = toml_file
    if not toml_file.exists():
        TOML.init()
    if pytoml_file.exists():
        pyproject_toml = FrozenDict(**TOML.read(pytoml_file))
    else:
        pyproject_toml = {}

    # Step[5]: TOML-Collect { Apps }
    toml_data = TOML.read()
    self.toml = FrozenDict(**toml_data.get("spoc", {}))
    self.apps = self.toml.get("apps", {})
    self.app_mode = self.toml.get("config", {}).get("mode", "development")
    self.pyproject = pyproject_toml

    # Step[6]: Load { Project }
    try:
        import project

        self.project = project
    except ImportError as exception:
        raise ValueError("Missing { project } module.") from exception

    # Step[7]: Load { Settings }
    if not hasattr(self.project, "settings"):
        raise ValueError("Missing { project.settings } module.")
    self.settings = self.project.settings

    # Step[8]: Collect { Apps }
    self.installed_apps = collect_installed_apps(self)

    # Finally: Collect { Keys }
    self.keys = [x for x in get_fields(self) if x not in ["init", "load_apps"]]

    # Step[9]: Create <class: Project>
    create_project(self, self.__theclass__, self.base_dir)

    # Step[6]: Load { Project }
    try:
        import project

        self.project = project
    except ImportError as exception:
        raise ValueError("Missing { project } module.") from exception

    self.core = Project(**{k: None for k in PROJECT_KEYS})

def load_apps(self, installed_apps: list[str] = None):
    """Load { Apps }"""

    if installed_apps and isinstance(installed_apps, list | set):
        # Discovering { Plugin(s) }
        loaded_plugins = get_plugins(self.__plugins__, installed_apps)
        app_schema = get_spoc_plugins(loaded_plugins.plugins)

        # Self { Definitions }
        self.schema = app_schema.schema
        self.modules = app_schema.modules
        self.plugins = loaded_plugins


def create_singleton(name: str, plugins: list[str]):
    """Base { Singleton }"""

    custom_class = type(
        name,
        (FrameworkSingleton,),
        {
            "init": functools.partial(init, plugins=plugins),
            "load_apps": load_apps,
        },
    )

    return custom_class


def framework(cls):
    """Create { Singleton }"""

    module = cls.__module__
    if cls.__module__ == "__main__":
        module = "main"

    name = cls.__name__

    if not hasattr(cls, "plugins"):
        raise Exception(
            f"Class <{module}.{name}> Must Define What <Plugins> To Look For."
        )

    plugins = cls.plugins

    return create_singleton(name=name, plugins=plugins)



"""
__________               __________                   __               __   
\______   \ ___________  \______   \_______  ____    |__| ____   _____/  |_ 
 |     ___// __ \_  __ \  |     ___/\_  __ \/  _ \   |  |/ __ \_/ ___\   __\
 |    |   \  ___/|  | \/  |    |     |  | \(  <_> )  |  \  ___/\  \___|  |  
 |____|    \___  >__|     |____|     |__|   \____/\__|  |\___  >\___  >__|  
               \/                                \______|    \/     \/      
"""


from fastberry import Project

import spoc

# Base Directory
BASE_DIR = spoc.root(__file__)[0]

# App
App = Project(base_dir=BASE_DIR, mode="cli")

"""
___________              __   
\__    ___/___   _______/  |_ 
  |    |_/ __ \ /  ___/\   __\
  |    |\  ___/ \___ \  |  |  
  |____| \___  >____  > |__|  
             \/     \/        
"""

# print(App.admin.keys)
# print(App.admin.modules["types.app_two.mytype"])
# print(Project().admin.plugins)
# print(Project().admin.schema)
# print(Project().admin.schema.types["app_two.mytype"])
# print(Project().admin.schema)
# print(App.admin.keys)

#print(App.core.modules)
print(App.core.keys)
# print(App.admin.plugins.modules)
"""
print(App.core.schema.types)
print(App.core.schema.graphql)
print(App.core.schema.router)
print(App.core.schema.commands)
"""



"""
    Components
"""
import functools

from .types import Component


# Function
def component(
    cls: object = None,
    *,
    config: dict = None,
    metadata: dict = None,
):
    """Plugin Creator"""
    config = config or {}
    metadata = metadata or {}
    if cls is None:
        return functools.partial(
            component,
            config=config,
            metadata=metadata,
        )

    # Real Wrapper
    cls.__spoc__ = Component(config=config, metadata=metadata)
    return cls
